package experimentalcode.tobias._index.sstree;


import java.util.ArrayList;
import java.util.List;
import java.util.PriorityQueue;

public class SSDataNode<T extends DoubleVector> extends SSNode<T>{

	private List<T> data;
	public boolean alreadyReinserted;
	
	public SSDataNode(SSTree sstree, boolean isRoot, SSDirectoryNode parent, int m, int M, NodeEntry accomdatingEntry){
		super(sstree, isRoot, parent, m, M, accomdatingEntry);
		data = new ArrayList<T>();
		alreadyReinserted = false;
	}
	@Override
  public void insert(T dv){
		this.data.add(dv);
		this.update(true);
		
	}
	
	public List<T> getData(){
		return data;
	}
	/**
	 * @return number of ChildNodes if it is a directory node. Number of datapoints if it is a leave.
	 */
	@Override
  public int size(){
		return data.size();
	}

	/**
	 * Updates the centroid, radius and checks if a split is necessary
	 *
	 */
	@Override
  public void update(boolean recursive){
		if(data.size() > M){
			if(alreadyReinserted || isRoot())
				split();
			else
				reinsert();
		}else{
			int points = 0;
			DoubleVector centroid = calcCentroid(data);
			double radius = 0.0;
			for(T dv:data){
				radius = Math.max(radius, df.distance(centroid, dv).getValue());
				points++;
			}
			accomdatingEntry.updateCR(centroid, radius);
			accomdatingEntry.totalDataPoints = points;
			if(!isRoot()&&recursive)parent.update(true);
		}
	}
	/**
	 * finds the dimension with the highest variance
	 *
	 */
	private void split(){
		//1. Find the dimension with the highest variance
		ArrayList<T> toSplit = new ArrayList<T>();
		for(T dataPoint:data)
			toSplit.add(dataPoint);
		int maxVarDim = calcSplitAxis(toSplit);
		
		//2. Sort points according to split-dimension
		PriorityQueue<DistanceObjectTuple> sortedPoints = new PriorityQueue<DistanceObjectTuple>();
		for(T dataPoint:data)
			sortedPoints.add(new DistanceObjectTuple(dataPoint.getValue(maxVarDim+1), dataPoint));
		ArrayList<T> sortedList = new ArrayList<T>();
		while(!sortedPoints.isEmpty())
			sortedList.add((T)sortedPoints.poll().object);
		//3. Find the best splitting Index
		int bestSplitIndex = splitIndex(maxVarDim, sortedList);
		//4. Then split according to the best splitting found
		List<T> node1 = getSubList(0, bestSplitIndex, sortedList);
		List<T> node2 = getSubList(bestSplitIndex, sortedList.size(), sortedList);
		if(isRoot()){
			SSDirectoryNode newRoot = new SSDirectoryNode(sstree, true,null, m, M, new NodeEntry());
			newRoot.accomdatingEntry.ssnode = newRoot;
			this.isRoot = false;
			NodeEntry child1 = newRoot.addNodeEntry();
			NodeEntry child2 = newRoot.addNodeEntry();
			
			this.accomdatingEntry = child1;
			this.parent = newRoot;
			SSDataNode newNode = new SSDataNode(sstree, false, newRoot, m, M, child2);
			
			this.data = node1;
			newNode.data = node2;
			
			child1.setSSNode(this);
			child2.setSSNode(newNode);
			
			sstree.root = newRoot;
			this.update(false);
			this.alreadyReinserted = false;
			newNode.update(false);
			newNode.alreadyReinserted = false;
			parent.update(true);
		}else{
			//remove data from old Node
			data = node1;
			NodeEntry newPage = parent.addNodeEntry();
			SSDataNode newNode = new SSDataNode(sstree, false, parent, m, M, newPage);
			newPage.setSSNode(newNode);
			newNode.data = node2;

			this.update(false);
			newNode.update(false);
			parent.update(true);
		}
	}
	/**
	 * reinserts 30% of the points in the page - not very efficcent
	 *
	 */
	private void reinsert(){
		int childsToReinsert = (int)Math.ceil((0.3*data.size()));
		if(childsToReinsert==0)return;
		PriorityQueue<DistanceObjectTuple> pq = new PriorityQueue<DistanceObjectTuple>(childsToReinsert, new DistanceObjectTuple.DesendingComparator());
		for(T dataPoint: data){
			pq.add(new DistanceObjectTuple((df.distance(dataPoint, this.accomdatingEntry.centroid).getValue()), dataPoint));
			if(pq.size()>childsToReinsert)pq.poll();
		}
		double maxDist = pq.peek().distance;
		List<T> reinsertList = new ArrayList<T>();
		for(int i = 0; i<data.size();i++){
			T dataPoint = data.get(i);
			if(df.distance(dataPoint, this.accomdatingEntry.centroid).getValue() > maxDist){
				reinsertList.add(data.remove(i));
				i--;
			}
		}
		this.alreadyReinserted=true;
		this.update(true);
		for(T dataPoint:reinsertList){
			this.sstree.insert(dataPoint);
		}
	}
}
