package experimentalcode.tobias._index.sstree;


import java.util.ArrayList;
import java.util.List;

import de.lmu.ifi.dbs.elki.distance.distancefunction.EuclideanDistanceFunction;



public abstract class SSNode<T extends DoubleVector> {
	
	SSDirectoryNode parent;
	boolean isRoot, isDataNode;
	public int m, M;
	EuclideanDistanceFunction df;
	public NodeEntry accomdatingEntry;
	public SSTree sstree;
	
	protected SSNode(SSTree sstree, boolean isRoot, SSDirectoryNode parent, int m, int M, NodeEntry accomdatingEntry){
		this.isRoot = isRoot;
		this.m = m;
		this.M = M;
		df = EuclideanDistanceFunction.STATIC;
		this.parent = parent;
		this.accomdatingEntry = accomdatingEntry;
		this.sstree=sstree;
	}
	
	public abstract void insert(T dv);
	public abstract void update(boolean recursive);
	public abstract int size();
	
	/**
	 * @return true if this node is the root, false otherwise
	 */
	public boolean isRoot(){
		return isRoot;
	}
	protected double calcVariance(List<T> dataNode){
		double variance = 0.0;
		DoubleVector c = calcCentroid(dataNode);
		for(T dv:dataNode){
			double dist = df.distance(dv, c).getValue();
			variance += dist*dist;
		}
		return variance;
	}
	protected double calcMaxDist(List<T> dvs,T centroid){
		double maxRad = 0.0;
		
		for(T dv:dvs)
			maxRad = Math.max(maxRad, df.distance(dv, centroid).getValue());
		return maxRad;
	}
	protected DoubleVector calcCentroid(List<T> dvs){
		double[] centroid = new double[dvs.get(0).getDimensionality()];
		for(T dv:dvs){
			for(int d=0; d<centroid.length;d++){
				centroid[d]+=dv.getValue(d+1);
			}
		}
		for(int d=0; d<centroid.length;d++)
			centroid[d]/=dvs.size();
		return new DoubleVector(centroid);
	}
	/**
	 * Calculates the dimension with the highest variance
	 * @param set the set of FeatureVectors to be split
	 * @return the dimension with the highest variance
	 */
	protected int calcSplitAxis(List<T> set){
		double[] sum = new double[set.get(0).getDimensionality()];
		double[] quadSum = new double[set.get(0).getDimensionality()];
		for(T dv: set){
			for(int d = 0; d<dv.getDimensionality();d++)
			{
				sum[d] += dv.getValue(d+1);
				quadSum[d] += dv.getValue(d+1)*dv.getValue(d+1);
			}
		}
		double maxVar = 0.0;
		int maxVarDim =0;
		for(int d = 0; d<sum.length;d++)
		{
			double var =quadSum[d]/set.size() - (sum[d]/set.size())*(sum[d]/set.size());
			if(var > maxVar){
				maxVar = var;
				maxVarDim = d;
			}
		}
		//if(maxVarDim == -1)
			//System.out.println("asdf");
		return maxVarDim;
	}
	/**
	 * Finds the index of the array to split the set in two sets with minimum sum of variance
	 * @param dimension the dimension in which the set should be split
	 * @param sortedList the set
	 * @return the index of the best splitpositino in the array
	 */
	protected int splitIndex(int dimension, List<T> sortedList){
		double minSumVariance = Double.MAX_VALUE;
		int bestSplitIndex = m;
		for(int splitIndex = m; splitIndex <= M-m+1;splitIndex++){
			List<T> page1 = sortedList.subList(0, splitIndex);
			List<T> page2 = sortedList.subList(splitIndex, sortedList.size());
			double sumVariance = calcVariance(page1)+calcVariance(page2);
			if(sumVariance < minSumVariance){
				bestSplitIndex = splitIndex;
				minSumVariance = sumVariance;
			}
		}
		return bestSplitIndex;
	}
	protected <T> List<T> getSubList(int start, int end, List<T> input){
		ArrayList<T> result = new ArrayList<T>();
		for(int i = start; i<end;i++)
			result.add(input.get(i));
		return result;
	}
	@Override
  public String toString(){
		if(isRoot())return "Root";
		String out = "(radius:"+accomdatingEntry.radius+" - totalData: "+accomdatingEntry.totalDataPoints+")";
		return out;
	}
}














